// DeloneTriangulation.cc

#include <sstream>
#include <cstdlib>
#include <cmath>
#include <gmp.h>

#include "Voronoi.h"
#include "DeloneTriangulation.h"
#include "la_support.h"
#include "lrslib.h"

SimplexContainer DeloneTriangulation::all_simplices;

//
//
//

void die_delone_triangulation(const char* last_words)
{
  std::cout <<"DeloneTriangulation.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

//
//
//

DeloneTriangulation::DeloneTriangulation()
  : simplices(optimal_hash_table_size)
{
  no_neighbours = 0;
}

DeloneTriangulation::DeloneTriangulation(const DeloneTriangulation& Del)
  : simplices(optimal_hash_table_size)
{
  simplices = Del.simplices;
  hash_simplices::iterator pos = simplices.begin();
  while (pos != simplices.end()) {
    all_simplices.add_simplex(*(*pos));
    pos++;
  }
  no_neighbours = 0;
}

//
//
//

DeloneTriangulation& DeloneTriangulation::operator=(const DeloneTriangulation& Del)
{
  if (this != &Del) {

    {
      destroy_neighbours();
      hash_simplices::iterator pos = simplices.begin();
      while (pos != simplices.end()) {
	all_simplices.delete_simplex(*(*pos));
	pos++;
      }
    }

    simplices = Del.simplices;
    hash_simplices::iterator pos = simplices.begin();
    while (pos != simplices.end()) {
      all_simplices.add_simplex(*(*pos));
      pos++;
    }
    no_neighbours = 0;
  }
  return *this;
}

//
//
//

DeloneTriangulation::~DeloneTriangulation()
{
  destroy_neighbours();
  hash_simplices::iterator pos = simplices.begin();
  while (pos != simplices.end()) {
    all_simplices.delete_simplex(*(*pos));
    pos++;
  }
}

//
//
//

int DeloneTriangulation::compute_no_neighbours()
{
  if (no_neighbours == 0) 
    make_neighbours();
  return no_neighbours;
}

//
//
//

DeloneTriangulation DeloneTriangulation::compute_nth_neighbour(const int n)
{
  if (no_neighbours == 0)
    die_delone_triangulation("no_neighbours == 0.\n");
  if (!(n >= 0 && n <= no_neighbours))
    die_delone_triangulation("!(n >= 0 && n < no_neighbours).\n");
  return do_flip((*irredundants[n]).facets);
}

//
//
//

void DeloneTriangulation::read_COOP_file(FILE* coop_file)
{
  // dimension
  
  int dimension;
  fscanf(coop_file, "%d", &dimension);
  if (dim != dimension)
    die_delone_triangulation("dim != dimension.\n");

  // simplices
  
  int half_simplices_size;
  fscanf(coop_file, "%d", &half_simplices_size);

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  for (int simplex_count = 0; simplex_count < half_simplices_size; simplex_count++) {
    for (int i = 0; i < dim+1; i++)
      vertices[0][i] = 0.0;
    for (int i = 0; i < dim+1; i++)
      vertices[i][dim] = 1.0;
    for (int i = 0; i < dim; i++)
      for (int j = 0; j < dim; j++)
	fscanf(coop_file, "%lf", &vertices[i+1][j]);

    Simplex L(vertices);
    Simplex* L1 = all_simplices.add_simplex(L);
    simplices.insert(L1);
    Simplex Lneg = L.compute_negative();
    Simplex* L2 = all_simplices.add_simplex(Lneg);
    simplices.insert(L2); 
  }

  doubleDestroySquareMatrix(dim+1, &vertices);

  // do not care about the rest
}

void DeloneTriangulation::write_COOP_file(FILE* coop_file) const
{
  // dimension
  fprintf(coop_file, "%d\n\n", dim);

  // simplices
  std::set<Simplex> half_simplices;
  hash_simplices::iterator simplices_pos = simplices.begin();
  while (simplices_pos != simplices.end()) {
    if (half_simplices.find(*(*simplices_pos)) == half_simplices.end() &&
	half_simplices.find((*(*simplices_pos)).compute_negative()) == half_simplices.end())
      half_simplices.insert(*(*simplices_pos));
    simplices_pos++;
  }
  fprintf(coop_file, "%d\n", half_simplices.size());
  for (std::set<Simplex>::iterator pos = half_simplices.begin(); pos != half_simplices.end(); pos++)
    (*pos).print_me(coop_file);

  // basis
  fprintf(coop_file, "%d\n", dim*(dim+1)/2);
  for (int i = 0; i < dim; i++)
    for (int j = 0; j <= i; j++)
      for (int row = 0; row < dim; row++) {
	for (int col = 0; col <= row; col++) 
	  if (row == i && col == j) fprintf(coop_file, "1 ");
	  else fprintf(coop_file, "0 ");
	fprintf(coop_file, "\n");
      }
  fprintf(coop_file, "\n");

  // inequality
  fprintf(coop_file, "%d\n", irredundants.size());
  for (unsigned int i = 0; i < irredundants.size(); i++) {
    for (int j = 0; j < dim*(dim+1)/2; j++)
      fprintf(coop_file, "%d ", int((*irredundants[i]).q[j]));
    fprintf(coop_file, "\n");
  }

  // max_iterations
  fprintf(coop_file, "\n500\n");

  // duality_gap
  fprintf(coop_file, "\n1e-10\n");
}

//
// compute_lower_bound
// see "Computational Approaches", Sec.10: Moments of Inertia
//

void DeloneTriangulation::compute_lower_bound(NTL::ZZ& invariant, double& lower_bound)
{
  double** C;
  doubleMakeSquareMatrix(dim, &C);
  hash_simplices::iterator pos = simplices.begin();
  while (pos != simplices.end()) {
    (*(*pos)).inc_edge_matrix(C);
    pos++;
  }

  // compute invariant

  NTL::mat_ZZ zz_C;
  zz_C.SetDims(dim,dim);
  for (int i = 0; i < dim; i++)
    for (int j = 0; j < dim; j++)
      zz_C[i][j] = NTL::to_ZZ(C[i][j]);

  determinant(invariant, zz_C, true);

  // compute lower_bound

  double q = double(simplices.size());
  double n = double(dim);
  double m = n/(q*(n+1)*(n+1));
  for (int i = 0; i < dim; i++)
    doubleMultiplyVector(dim, m, C[i], C[i]);
  
  determinant(dim, C, &lower_bound);    
  lower_bound = sqrt(fabs(lower_bound));

  doubleDestroySquareMatrix(dim, &C);
}

//
//
//

int permutation_sign(int *pi)
{
  int sign = 1;
  for (int i = 0; i < dim; i++)
    for (int j = i+1; j < dim; j++)
      if (pi[i] > pi[j]) sign = -sign;
  return sign;
}

//
//
//

bool check_equivalence(const hash_simplices& simplices0, const hash_simplices& simplices1, double** linear_map)
{
  bool equivalent = true;
  hash_simplices::iterator L_pos = simplices0.begin();
  Simplex L = *(*L_pos); // only to safe some constructors and destructors
  while (L_pos != simplices0.end() && equivalent) {
    L = (*(*L_pos)).perform_linear_map(linear_map);
    equivalent = (simplices1.find(&L) != simplices1.end());
    L_pos++;
  }
  return equivalent;
}

//
//
//

bool DeloneTriangulation::is_equivalent(DeloneTriangulation& Del) 
{
  // 1. check some invariants

  if (simplices.size() != Del.simplices.size())
    return false;

  NTL::ZZ invariant1, invariant2;
  double lower_bound1, lower_bound2;
  compute_lower_bound(invariant1, lower_bound1);

  Del.compute_lower_bound(invariant2, lower_bound2);

  if (invariant1 != invariant2)
    return false;
  
  // 2. work hard

  bool equivalent = false;

  double** linear_map;
  doubleMakeSquareMatrix(dim, &linear_map);
  double** inverse;
  doubleMakeSquareMatrix(dim, &inverse);
  double inv_det;
  int* pi;
  intMakeVector(dim, &pi);
  
  Simplex startL = *(*simplices.begin());
  Simplex Lnew = startL; // safe constructor and destructor
  startL.compute_inverse(inverse, inv_det);
  hash_simplices::iterator pos = Del.simplices.begin(); 
  while (pos != Del.simplices.end() && !equivalent) {
    for (int n = 0; n < dim+1 && !equivalent; n++) {
      for (int i = 0; i < dim; i++)
	pi[i] = i;
      do {
	if (!(*(pi) < 0)) {  // replacement for "(sign(pi) == 1) {"
	  if (startL.find_linear_map(*(*pos), n, pi, inverse, inv_det, linear_map)) {
	    Lnew = startL.perform_linear_map(linear_map);
	    if (Del.simplices.find(&Lnew) == Del.simplices.end()) 
	      die_delone_triangulation("Del.simplices.find(&Lnew) == Del.simplices.end().\n");
	    
	    equivalent = check_equivalence(simplices, Del.simplices, linear_map);
	  }
	  else
	    equivalent = false;
	}
      }
      while (!equivalent && std::next_permutation(&pi[0], &pi[dim]));
    }
    pos++;
  }

  doubleDestroySquareMatrix(dim, &inverse);
  doubleDestroySquareMatrix(dim, &linear_map);
  intDestroyVector(dim, &pi);

  return equivalent;
}

//
// 
//

DeloneTriangulation principal_domain()
{
  DeloneTriangulation Del;

  double** frame;
  doubleMakeSquareMatrix(dim+1, &frame);
  doubleSetIdentity(dim+1, frame);
  for (int i = 0; i < dim; i++)
    frame[dim][i] = -1.0;
  for (int i = 0; i < dim+1; i++)
    frame[i][dim] = 1.0;

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  do {
    for (int i = 0; i < dim+1; i++)
      if (i != 0) {
	doubleCopyVector(dim+1, vertices[i-1], vertices[i]);
	doubleAddVector(dim, 1.0, frame[i], vertices[i]);
      }
      else
	doubleCopyVector(dim+1, frame[0], vertices[0]);
    Simplex L(vertices);
    // here we add too many simplices to all_simplices
    Simplex* L1 = &(const_cast<Simplex&>(*Del.all_simplices.add_simplex(L)));
    Del.simplices.insert(L1);
  }
  while (std::next_permutation(&(frame[0]), &(frame[dim+1])));

  doubleDestroySquareMatrix(dim+1, &vertices);
  doubleDestroySquareMatrix(dim+1, &frame);

  return Del;
}

//
//
//

DeloneTriangulation barnes_trenerry_domain()
{
  DeloneTriangulation Del;

  int *pi;
  intMakeVector(dim, &pi);
  for (int i = 0; i < dim; i++)
    pi[i] = i;

  double** frame;
  doubleMakeSquareMatrix(dim, &frame);
  doubleSetIdentity(dim, frame);

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  int m = (dim-1)/2;

  do {
    
    for (int i = 0; i < dim+1; i++)
      for (int j = 0; j < dim; j++)
	vertices[i][j] = 0.0;
    for (int i = 0; i < dim+1; i++)
      vertices[i][dim] = 1.0;

    int vertex_pos = 1;

    for (int r = 1; r <= m; r++) {
      for (int i = 0; i < r; i++)
	doubleAddVector(dim, 1.0, frame[pi[i]], vertices[vertex_pos]);
      vertex_pos++;
    }
    for (int s = 0; s <= m; s++) {
      for (int i = 1; i <= dim; i++)
	if (i <= m+s)
	  doubleAddVector(dim, 0.5, frame[pi[i-1]], vertices[vertex_pos]);
	else
	  doubleAddVector(dim, -0.5, frame[pi[i-1]], vertices[vertex_pos]);
      vertex_pos++;
    }

    // apply transformation T
    for (int i = 1; i < dim+1; i++) {
      double y1 = vertices[i][0];
      vertices[i][0] = y1 + vertices[i][1];
      for (int j = 1; j < dim; j++)
	vertices[i][j] = y1 - vertices[i][j];
    }

    Simplex L(vertices);
    // here we add too many simplices to all_simplices
    Simplex* L1 = &(const_cast<Simplex&>(*Del.all_simplices.add_simplex(L)));
    Del.simplices.insert(L1);
  }
  while (std::next_permutation(&(pi[0]), &(pi[dim])));

  doubleDestroySquareMatrix(dim+1, &vertices);
  doubleDestroySquareMatrix(dim, &frame);
  intDestroyVector(dim, &pi);

  return Del;
}

//
//
//

void DeloneTriangulation::make_neighbours()
{
  if (no_neighbours != 0)
    die_delone_triangulation("second call of make_neighbours().\n");

  make_facets_and_regulators();

  // consistency check
  if (regulators.size() < unsigned(Dim))
    die_delone_triangulation("regulators.size() < unsigned(Dim).\n");

  // lrs

  if (!lrs_init(" * computing irredundant facets : "))
    die_delone_triangulation("lrs_init failed.\n");
      
  lrs_dat *Q;
  Q = lrs_alloc_dat("LRS globals");
  if (Q == NULL)
    die_delone_triangulation("lrs_alloc_dat failed.\n");
  Q->m = long(regulators.size());
  Q->n = long(Dim+1);
  Q->hull = TRUE;

  lrs_dic *P;
  P = lrs_alloc_dic(Q);
  if (P == NULL)
    die_delone_triangulation("lrs_alloc_dic failed.\n");

  // inequalities
  
  long num[Dim+1];
  long den[Dim+1];
  for (int i = 0; i < Dim+1; i++)
    den[i] = 1L;
  num[0] = 0L;
  
  long row = 1L;
  std::vector <std::set<Regulator>::iterator> indices(regulators.size());
  std::set<Regulator>::iterator reg_pos = regulators.begin();
  while (reg_pos != regulators.end()) {
    for (long i = 1; i < Dim+1; i++) 
      num[i] = long((*reg_pos).q[i-1]);
    lrs_set_row(P, Q, row, num, den, GE);
    indices[row-1] = reg_pos;
    
    row++;
    reg_pos++;
  }

  lrs_mp_matrix Lin;
  
  itomp(ONE, Q->Lcm[0]); // circumvent a bug in lrs

  if (!lrs_getfirstbasis(&P, Q, &Lin, TRUE)) // TRUE means no_output
    die_delone_triangulation("lrs_getfirstbasis failed.\n");

  long m = P->m_A;
  long d = P->d;
  long nlinearity = Q->nlinearity;
  long lastdv = Q->lastdv;
  long *redineq = new long[m+1];
  long ineq;

  for (long i = 0; i < nlinearity; i++)
    redineq[Q->linearity[i]] = 2L;

  for (long index = lastdv + 1; index <= m + d; index++) {
    ineq = Q->inequality[index - lastdv];
    redineq[ineq] = checkindex (P, Q, index);
    if (redineq[ineq] != ONE) {
      irredundants.push_back(indices[ineq-1]);
    }
  }		   

  lrs_free_dic(P, Q);
  lrs_free_dat(Q);

  delete [] redineq;

  lrs_close(" * -> finished : ");

  // lrs end

  no_neighbours = irredundants.size();
}

//
//
//

void DeloneTriangulation::destroy_neighbours()
{
  if (no_neighbours != 0) {
    irredundants.clear();
    destroy_facets_and_regulators();
    no_neighbours = 0;
  }
}

//
//
//

void DeloneTriangulation::make_facets_and_regulators()
{
  // make facets

  double** facet;
  doubleMakeSquareMatrix(dim+1, &facet);

  hash_simplices::iterator pos = simplices.begin();  
  int balance = 0;

  while (pos != simplices.end()) {
    for (int n = 0; n < dim+1; n++) {
      (*(*pos)).get_nth_facet(n, facet);
      std::pair<std::set<Facet>::iterator, bool> result = facets.insert(Facet(facet));
      if (!result.second) {
	const_cast<Facet&>(*result.first).insert_out_vertex(facet[dim]);
	balance--;
      }
      else
	balance++;
    }
    pos++;
  }
  doubleDestroySquareMatrix(dim+1, &facet);  

  if (balance != 0) 
    die_delone_triangulation("balance != 0.\n");

  // make regulators

  std::set<Facet>::iterator facet_pos = facets.begin();
  while (facet_pos != facets.end()) {
    bool repartition = true;

    double **vertices;
    doubleMakeSquareMatrix(dim+1, &vertices);
    for (int i = 0; i < dim+1; i++) {
      if ((*facet_pos).alpha[i] < 0.0) {
	doubleCopySquareMatrix(dim+1, (*facet_pos).in_simplex, vertices);
	doubleCopyVector(dim+1, (*facet_pos).out_vertex, vertices[i]);
	Simplex Lnew(vertices);
	if (repartition) 
	  repartition = (simplices.find(&Lnew) != simplices.end());
      }
    }
    doubleDestroySquareMatrix(dim+1, &vertices);

    if (repartition) {
      std::pair<std::set<Regulator>::iterator, bool> result = regulators.insert(Regulator(*facet_pos));
      const_cast<Regulator&>(*result.first).insert_facet(&(*facet_pos));
    }
    facet_pos++;
  }
}

//
//
//

void DeloneTriangulation::destroy_facets_and_regulators()
{
  facets.clear();
  regulators.clear();
}

//
//
//

DeloneTriangulation DeloneTriangulation::do_flip(std::list<Facet*> flip_facets) const
{
  DeloneTriangulation Del;
  Del.simplices = simplices;
  hash_simplices::iterator pos = Del.simplices.begin();
  while (pos != Del.simplices.end()) {
    all_simplices.add_simplex(*(*pos));
    pos++;
  }

  double** vertices;
  doubleMakeSquareMatrix(dim+1, &vertices);

  while (!flip_facets.empty()) {
    Facet& facet = *flip_facets.front();

    // consistency check
    if (!(facet.alpha[dim] < 0.0))
      die_delone_triangulation("alpha[dim] >= 0.0.\n");
    
    for (int i = 0; i < dim+1; i++) {
      doubleCopySquareMatrix(dim+1, facet.in_simplex, vertices);
      doubleCopyVector(dim+1, facet.out_vertex, vertices[i]);

      Simplex L(vertices);
      if (facet.alpha[i] > 0.0) {
	// flip in vertex[i]
	Simplex* L1 = &(const_cast<Simplex&>(*Del.all_simplices.add_simplex(L)));
	Del.simplices.insert(L1);
      }
      if (facet.alpha[i] < 0.0) {
	// flip out vertex[i]
	Del.simplices.erase(&L);
      }
    }
    // flip out in_simplex
    Simplex L(facet.in_simplex);
    Del.simplices.erase(&L);
    flip_facets.pop_front();
  }

  doubleDestroySquareMatrix(dim+1, &vertices);

  // consistency check
  if (simplices.size() > unsigned(factorial(dim)))
    die_delone_triangulation("simplices.size() > factorial(dim).\n");
  
  return Del;
}

