// Simplex.h

#ifndef _SIMPLEX_H
#define _SIMPLEX_H

#include <cstdlib>

// every d-dimensional simplex L = conv{v0, v1, ..., vd} is stored 
// as L' = conv{0, w0, ..., wd} where 0 < w0 < ... < wd (lexicographic order)
// and L = L' + v with v \in Z^d.

// We use homogeneous coordinates where 1.0 appears last 

class Simplex {
public:
  Simplex(double** v);
  Simplex(const Simplex& L);
  ~Simplex();

  Simplex compute_negative() const;

  void get_nth_facet(const int n, double** facet) const;
  // for L = conv{0, v1, ..., vd} is the n-th facet given by
  // n = 0: F = (0, v2-v1, ..., vd-v1 | -v1)
  // n > 0: F = (0, v1, ..., vn-1, vn+1, ..., vd | vn)
  // facet is a (d+1)*(d+1)-matrix

  void compute_inverse(double **inv_vertices, double& inv_det) const;
  bool find_linear_map(const Simplex& L, const int n, const int* pi, 
		       double** inv_vertices, double inv_det, double** linear_map) const;
  // find the linear map (a d*d-matrix) which transforms this into L - v_n
  // where v_n is the n-th vertex of L. (false if the map is not in \GLdZ)
  Simplex perform_linear_map(double** linear_map) const;

  void inc_edge_matrix(double **edge_matrix) const;

  void compute_circumsphere_matrix(const int i, const int j, double *F) const;
  // gives the F_(i,j)-th component of the "circumsphere-matrix"

  size_t compute_hash_no() const;

  void print_me(FILE* f) const;

  Simplex& operator=(const Simplex& L);
  bool operator==(const Simplex& L) const;
  bool operator<(const Simplex& L) const;

private:
  double **vertices; // a (d+1)*(d+1)-matrix 
  size_t hash_no;
};


struct SimplexCmp
{
  bool operator()(const Simplex* L1, const Simplex* L2) const
  {
    return *L1 < *L2;
  }
};

struct SimplexEq
{
  bool operator()(const Simplex* L1, const Simplex* L2) const
  {
    return *L1 == *L2;
  }
};

struct SimplexHash
{
  size_t operator()(const Simplex *L) const
  {
    return (*L).compute_hash_no();
  }
};

#endif
