// scc.cc

#include <deque>
#include <vector>
#include <list>
#include <iostream>
#include <unistd.h>
#include <NTL/ZZ.h>

#include "Voronoi.h"
#include "DeloneTriangulation.h"
#include "la_support.h"

int dim;
int Dim;
int optimal_hash_table_size;

void die_scc(char* last_words)
{
  std::cout <<"scc.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

//

int main(int argc, char** argv)
{
  bool random_walk_option = false;
  bool greedy_walk_option = false;
  bool read_input_option = false;
  bool write_coop_option = false;

  int number_random_steps = 0;

  int c;
  dim = 0;

  while ((c = getopt (argc, argv, "hgiwd:r:")) != -1)
    switch (c)
      {
      case 'h':
	std::cout <<"Usage: scc [options] <file\n";
	std::cout <<"Options:\n";
	std::cout <<" -h  show this help\n";
	std::cout <<" -dD dimension D (required)\n";
	std::cout <<" -rN cruise randomly (N times)\n";
	std::cout <<" -g  cruise greedily\n";
	std::cout <<" -i  read coop file\n";
	std::cout <<" -w  write coop file for every secondary cone\n";
	return 0;
	break;
      case 'd':
	dim = atoi(optarg);
	Dim = dim*(dim+1)/2;
	optimal_hash_table_size = factorial(dim);
	break;
      case 'g':
	greedy_walk_option = true;
	break;
      case 'r':
	random_walk_option = true;
	number_random_steps = atoi(optarg);
	break;
      case 'i':
	read_input_option = true;
	break;
      case 'w':
	write_coop_option = true;
	break;
      default:
	die_scc("invalid option\nTry 'scc -h' for more information.\n");
	return 1;
      }

  if (dim <= 0)
    die_scc("invalid options\noption '-d' is needed.\n");

  std::cout <<"*** sc ***\n";
  std::cout <<"***  c ***\n\n";

  // prepare starting triangulation

  DeloneTriangulation Del;
  if (read_input_option) {
    Del.read_COOP_file(stdin);
  }
  else
    Del = principal_domain();
  
  std::vector<DeloneTriangulation> classification;
  std::list<int> known_adj; 
  std::deque< std::pair<int, std::list<int> > > todo;

  classification.push_back(Del);
  todo.push_back(make_pair(0, known_adj));

  //
  //
  //

  if (random_walk_option) {

    for (int current = 0; current < number_random_steps; current++) {
      NTL::ZZ invariant;
      double lower_bound;
      Del.compute_lower_bound(invariant, lower_bound);
      std::cout <<"Looking at #" <<current <<" (normalized covering density >= " <<lower_bound <<")\n";

      int no_neighbours = Del.compute_no_neighbours();
      std::cout <<" * There are " <<no_neighbours <<" neighbours.\n"; 

      if (write_coop_option) {
	std::cout <<" * Write output for COOP " <<std::endl;    
	FILE* coop_file;
	char file_name[100];
	sprintf(file_name, "sc%d_%d.coop", dim, current);
	coop_file = fopen(file_name, "w");
	Del.write_COOP_file(coop_file);
	fclose(coop_file);
      }

      int r = rand() % no_neighbours;
      std::cout <<" * Proceed with number " <<r <<".\n\n";
      DeloneTriangulation random_neighbour = Del.compute_nth_neighbour(r);
      Del = random_neighbour;
    }
  }

  else if (greedy_walk_option) {
    int current = 0;
    bool finished = false;

    while (!finished) {
      NTL::ZZ invariant;
      double lower_bound;
      Del.compute_lower_bound(invariant, lower_bound);
      std::cout <<"Looking at #" <<current <<" (normalized covering density >= " <<lower_bound <<")\n";

      int no_neighbours = Del.compute_no_neighbours();
      std::cout <<" * There are " <<no_neighbours <<" neighbours.\n"; 

      if (write_coop_option) {
	std::cout <<" * Write output for COOP " <<std::endl;    
	FILE* coop_file;
	char file_name[100];
	sprintf(file_name, "sc%d_%d.coop", dim, current);
	coop_file = fopen(file_name, "w");
	Del.write_COOP_file(coop_file);
	fclose(coop_file);
      }

      std::vector<DeloneTriangulation> neighbours(no_neighbours);
      for (int i = 0; i < no_neighbours; i++)
	neighbours[i] = Del.compute_nth_neighbour(i);
      
      int number_best = -1;
      Del.compute_lower_bound(invariant, lower_bound);
      double least_lower_bound = lower_bound;
      for (int i = 0; i < no_neighbours; i++) {
	neighbours[i].compute_lower_bound(invariant, lower_bound);
	if (lower_bound < least_lower_bound) {
	  least_lower_bound = lower_bound;
	  number_best = i;
	}
      }
      
      finished = (number_best == -1);
      if (!finished) {
	current++;
	Del = neighbours[number_best];
	std::cout <<"\n";
      }
    }
  }      

  else {

    while (!todo.empty()) {
      int current = todo.front().first;
      Del = classification[current];
      known_adj = todo.front().second;
      todo.pop_front();

      NTL::ZZ invariant;
      double lower_bound;
      Del.compute_lower_bound(invariant, lower_bound);
    
      std::cout <<"Looking at #" <<current <<" (normalized covering density >= " <<lower_bound <<")\n";
      std::cout <<" * Remaining: " <<todo.size() <<std::endl;    

      int no_neighbours = Del.compute_no_neighbours();

      if (write_coop_option) {
	std::cout <<" * Write output for COOP " <<std::endl;    
	FILE* coop_file;
	char file_name[100];
	sprintf(file_name, "sc%d_%d.coop", dim, current);
	coop_file = fopen(file_name, "w");
	Del.write_COOP_file(coop_file);
	fclose(coop_file);
      }

      std::vector<DeloneTriangulation> neighbours(no_neighbours);
      for (int i = 0; i < no_neighbours; i++)
	neighbours[i] = Del.compute_nth_neighbour(i);

      std::cout <<" * There are " <<no_neighbours <<" neighbours.\n"; 
      std::cout <<" * ";

      for (int i = 0; i < no_neighbours; i++) {

	std::cout <<"(" <<i <<std::flush;
	
	// 1. check if neighbour[i] = Del
	bool known = neighbours[i].is_equivalent(Del);
	if (known) 
	  std::cout <<"=" <<current <<") " <<std::flush;
	
	// 2. check if neighbour[i] = completed where adjacencies are known
	if (!known) {
	  // use gathered information for adjacency-tests to completed
	  std::list<int>::iterator pos = known_adj.begin(); 
	  while (pos != known_adj.end() && !known) {
	    known = neighbours[i].is_equivalent(classification[*pos]);
	    if (known) 
	      std::cout <<"=" <<*pos <<") " <<std::flush;
	    pos++;
	  }
	}

	// 3. check if neighbour[i] = todo[j]
	if (!known) {
	  for (unsigned int j = 0; j < todo.size() && !known; j++) {
	    known = neighbours[i].is_equivalent(classification[todo[j].first]);
	    if (known) {
	      // here we know that is todo[j] adjacent to Del (current)
	      if (find(todo[j].second.begin(), todo[j].second.end(), current) == todo[j].second.end()) 
		todo[j].second.push_back(current);
	      std::cout <<"=" <<todo[j].first <<") " <<std::flush;
	    }
	  }
	}

	// 4. All checks failed. Hence we have a new type.
	if (!known) {
	  // here we know at least that neighbours[i] is adjacent to Del.
	  std::list <int> new_adj;
	  new_adj.push_back(current);
	  classification.push_back(neighbours[i]);
	  todo.push_back(make_pair(classification.size()-1, new_adj));
	  std::cout <<"->" <<classification.size()-1 <<") " <<std::flush;
	}
      }
      std::cout <<"\n\n";
    }

    std::cout <<"--\n";
    std::cout <<"Classification completed!\n";
    std::cout <<"There are " <<classification.size() 
	      <<" non-equivalent Delone triangulations in dimension " <<dim <<".\n";
  }
  
  return 0;
}

